module Parser  where

import Char
import Exception
import AbsSyntax
import Subst
import Lex
import ParserMonad

infixr 1 `symThenP`
infixl 2 `chkSymP`

pair x y = (x,y)

p   `chkSymP` sym = p           `chkP`  symbolP sym
sym `symThenP` p  = symbolP sym `thenP` p

star f = optionalP [] (plus f)
plus f = return (:) `apP` f `apP` optionalP [] plus_lem
  where plus_lem = "," `symThenP` plus f

wrapP l r p = l `symThenP` p `chkSymP` r

inParenP = wrapP "(" ")"
inBracesP = wrapP "{" "}"
inBracketsP = wrapP "[" "]" -- 
inAngleBracketsP = wrapP "<" ">"

parseNames :: Parse [Name]
parseNames = star parseName

parseIden :: Parse String
parseIden = idenP

parseName :: Parse Name
parseName  = parseIden

parseBind :: Parse a -> Parse b -> Parse (a,b)
parseBind f g = return pair `apP` f `apP` g

parseBinds :: Parse a -> Parse b -> Parse [(a,b)]
parseBinds f g = star (parseBind f g)

parseDec :: Parse Exp
parseDec = ":" `symThenP` expP

parseNs = manyP parseName

{- Concrete syntax of expressions:

exp = exp bexp
    | bexp
    | \ names -> exp
    | let decls in exp
    | letrec decls in exp
    | ( context ) -> exp
    | { context } exp

bexp = aexp

aexp = name
     | con
     | sort
     | ( exp )

-}

expP, bexpP, aexpP :: Parse Exp

expP = 
        return absExp `chkSymP` "\\" `apP` parseNs
		      `chkSymP` "->" `apP` expP
  `orP` letsP
  `orP` useP
  `orP` return piExp `apP` inParenP (parseBinds parseName parseDec)
                    `chkSymP` "->"
	            `apP` expP
  `orP` return sigExp `apP` inBracesP (parseBinds parseName parseDec) 
                      `apP` expP
  `orP` return app `apP` someP bexpP
  where
    absExp = flip $ foldr mkAbs 
    piExp =  flip $ foldr (\(n,t1) -> mkPi n t1) 
    sigExp = flip $ foldr (\(n,t1) -> mkSig n t1)
    app = foldl1 mkApp

letsP = 
       absLet nonRecursive "let"
 `orP` absLet recursive "letrec"
 where
     absLet flag kwd = return (letExp flag) `chkSymP` kwd `apP` letbodyP
     letbodyP = return pair `apP` inBracesP letdefsP `chkSymP` "in" `apP` expP
     letExp f = uncurry (mkLet f)

-- fst version useP = return mkUse `chkSymP` "use" `apP` expP `chkSymP` ":" `apP` expP `chkSymP` "in" `apP` expP
useP = return mkUse `chkSymP` "use" `apP` expP `apP` coercP `chkSymP` "in" `apP` expP
  where
       coercP = return Coerc `chkSymP` ":" `apP` expP `orP` return Infer

pairsP = return (:) `apP` expP `chkSymP` "," `apP` plus expP

bexpP =
        return labelref `chkSymP` "#" `apP` idenP `apP` manyP labidentP
      `orP` return proj `apP` aexpP `apP` manyP ("." `symThenP` parseName)
      `orP` return mkFst `chkSymP` "fst" `apP` aexpP 
      `orP` return mkSnd `chkSymP` "snd" `apP` aexpP 
     where
        proj e sels = foldl ESel e sels
        labelref id lbs = foldl ESel (EIdent "dummylab") (id:lbs)

labidentP = "." `symThenP` parseName


aexpP =
        return mkIdent `apP` idenP 
  `orP` return (mkSort "Set") `chkSymP` "Set"
  `orP` return (mkSort "Type") `chkSymP` "Type"
  `orP` return (mkSort "Kind") `chkSymP` "Kind"
  `orP` return pairExp `apP` inBracesP (pairsP)
  `orP` return recTExp `apP` inAngleBracketsP parseRecT
  `orP` return recOExp `apP` inAngleBracketsP parseRecO
  `orP` inParenP expP
  where
    pairExp = foldr1 EPair
    recTExp (n,f) = if n == emptyRecT 
                    then  mkRecLabT emptyRecT f
                    else  mkRecLabT n f
 
    recOExp (n,f) = if n == emptyRecO
                    then  mkRecO emptyRecO f
                    else  mkRecO n f

mkRecLabT e s = let sl = labelSig s 0 in mkRecT e sl
    
labelSig [] _ = []
labelSig ((l,t):sig) i = 
         let ni = mkName i
             t1 = EAbs ni (substLab ni t)  
         in  (l,t1): labelSig sig (i+1)
         where mkName i = "rabs_" ++ show i
 

{- Concrete syntax of definitions 

   [rec] Name [Telescopes] : Type = Exp
-}

defsP = manyP (defP `chkSymP` ";")
letdefsP = return (:) `apP` defP `apP` manyP (";" `symThenP` defP)

defP :: Parse Def
defP =
-- previous wired           return (mkExpl True) `chkSymP` "rec" `apP` declaP
           return mkWired `chkSymP` "rec" `apP` dConsP
	 `orP` return (uncurry mkDefCons) `chkSymP` "cons" `apP` dConsP
         `orP` return (mkExpl False) `apP` declaP
	 where
	   mkExpl b (n,c,t,e) = -- inefficient!!!
	      let sig = concat c
                  typ = foldr (\(n,t1) -> mkPi n t1) t sig
	          exp = foldr EAbs e (map fst sig)
              in (if b then mkDefRec else mkDef) n exp typ
     
           varsOfPi acc (EPi x _ beta) = varsOfPi (acc++[x]) beta
	   varsOfPi acc _ = acc
           
           mkWired (n,t) = mkDefRec n (ECons n) t

		  
declaP :: Parse (Name,[[(Name,Exp)]],Exp,Exp)
declaP =
            return (,,,) `apP` idenP `apP` telescopesP `chkSymP` ":" 
            `apP`  expP `chkSymP` "=" `apP` expP

telescopesP = manyP telescopeP
telescopeP = inParenP (parseBinds idenP parseDec)

dConsP = return pair `apP` idenP `chkSymP` ":" `apP` expP


contextP = inBracketsP (parseBinds parseNum parseDec)

parseNum :: Parse Int
parseNum = return read `apP` numP 

   


-- Different parsers

parseExp = parse expP . lexanal
--parseJdge = parse judgeP . lexanal
parseDef = parse defP . lexanal
parseDefs = parse defsP . lexanal

-- additions for records

parseAssig :: Parse Exp
parseAssig = "=" `symThenP` expP

parseAssigs = parseBinds parseName parseAssig
parseFields = parseBinds parseName parseDec

parseRecO = 
        return pair `apP` expP `chkSymP` "," `apP` parseAssigs
  `orP` return (pair emptyRecO) `apP`  parseAssigs

parseRecT = 
        return pair `apP` expP `chkSymP` "," `apP` parseFields
  `orP` return (pair emptyRecT) `apP`  parseFields




{- bexp = bexp . ident
        | aexp 

aexp = aexp . ident
     | < fielddecl >
     | < fieldassig >
     | <>

fieldDecl = sigs
          | n,sigs
sigs = n1:t1,...,nm:tm

fieldAssig = assigs
           | n,assigs

assigs = n1=t1,...,nm=tm
-}
