module ParserMonad(
    Parse(..),         -- Parser monad type, should really be abstract
    --ErrorPos(..),
    parse,parse',      -- run a parser on some input
    apP,               -- standard monad operation, parse sequences
    thenP, chkP,       -- parse sequences, throws away one part
    orP,               -- parses alternatives
    manyP,someP,       -- repetion parsers (zero or more, resp. one or more)
    optionalP,         -- parser for optional items (zero or one)
    checkP, symbolP,   -- parsers that accept one terminal symbol
    idenP,             -- parser that accepts an identifier
    numP,              -- parser that accepts a number
    fileNameP,         -- parses filenames
    errorP,            -- parser that fails (not recommended)
    theRestP)          -- returns the remaining token stream
                       -- (e.g., to be able to split parsing into two stages)
    where

import Char
import AbsSyntax
import Lex
import Exception


infixr 0 `orP`
-- infixl 1 `bindP`,`thenP`
infixl 2 `apP`,`chkP`

-- Monad Parser 

data Parse a = Parse (ErrorPos -> [PosTOKEN] -> (ErrorPos,E (a,[PosTOKEN])))
mkParse = Parse
funOfParse (Parse f) = f

type ErrorPos = (Pos,TOKEN,[String]) -- high water mark

instance Monad Parse where
   return a = Parse (\ep tkl -> (ep, return (a,tkl)))
   (Parse p) >>= xp2 = 
               Parse (\ep tkl -> 
                      let (ep1,e1) = p ep tkl
                      in
                        elimE e1 (\(x,tkl2)-> funOfParse (xp2 x) ep1 tkl2)
                                 (\s-> (ep1,errorE s))
                     )

-- Parsing Combinators (abstract)
  
apP :: Parse (a->b) -> Parse a -> Parse b
fp `apP` xp = fp >>= \ f -> xp >>= \ x -> return (f x)

chkP :: Parse a -> Parse b -> Parse a
p1 `chkP` p2 = return const `apP` p1 `apP` p2

thenP :: Parse a -> Parse b -> Parse b
p1 `thenP` p2 = p1 >>= const p2

orP :: Parse a -> Parse a -> Parse a
p1 `orP` p2 = 
  Parse (\ ep tkl ->
           let (ep1,e1) = (funOfParse p1) ep tkl 
           in 
             elimE e1 (\xtkl1 -> (ep1,return xtkl1))
	              (\s -> let (ep2,e2) = (funOfParse p2) ep1 tkl
	                     in
                               elimE e2 (\xtkl2-> (ep2, return xtkl2))
			                (\s-> (ep2, errorE s))
                      )
        )

optionalP x p = p `orP` return x

manyP p = optionalP [] (someP p)
someP p = return (:) `apP` p `apP` manyP p

-- Parsing functions (lexer dependents)

-- Awful but necessary for uniformity of ParsCom

fileNameP :: Parse String
fileNameP = 
 mkParse (\ep tkl -> 
            (ep,return ((concatMap strOutOfTok  tkl),[eot])))
 where strOutOfTok (_,Iden s) = s
       strOutOfTok (_,Symbol "EOT") = []
       strOutOfTok (_,Symbol s) = s


theRestP :: Parse [PosTOKEN]
theRestP = mkParse (\ep tkl -> (ep,return (tkl,[eot])))

pickP :: Parse TOKEN
pickP = 
 mkParse (\ep tkl ->
              case tkl of
                ((_,t):_) -> (ep, return (t,tkl))
		tkl1 -> funOfParse (errorP "Unexpected end of file") ep tkl1
                ) 

skipP :: a -> Parse a
skipP x = mkParse (\ep tkl -> 
                    case tkl of
                      (_:tkl1) -> (ep,return (x,tkl1))
                      _ ->  funOfParse (errorP "Unexpected end of file") ep tkl
                  )

eofP = checkP (Symbol "EOT")

checkP t =
  pickP >>= \ t' ->
  if t'== t
  then skipP ()
  else errorP (string_of_token t)

symbolP = checkP . Symbol
       
idenP =
  pickP >>= \ t ->
  case t of
    Iden s -> skipP s
    _ -> errorP ("<identifier>")

numP =
  pickP >>= \ t ->
  case t of
    Iden s -> if all isDigit s
              then skipP s
              else errorP ("<number>")
    _ -> errorP ("<number>")

{-
fileNameP = 
 return fn `apP` (idenP `chkP` symbolP "." `chkP` checkExtP)
 where fn n = n ++ ".sr"

checkExtP = 
   pickP >>= \t -> 
   case t of 
       Iden s -> if s == "sr" then skipP ()
                 else errorP "extension sr"
       _ -> errorP "extension sr"
-}

-- Error management

errorP :: String -> Parse a
errorP msg = mkParse(\ep tkl ->
                       case tkl of
                           [] -> (updErr ep (eot,msg),errorE msg) -- hmm...
                           (pt:_) -> (updErr ep (pt,msg),errorE msg)
                    )


showErrPos :: (Pos,String) -> String
showErrPos (pos,msg) = "At " ++ showPos pos ++": "++msg
  where
    showPos pos@(line,column) =
      if pos==eotPos
      then "end of file"
      else unwords ["Line",show line,"column",show column]
  
showErrPos' :: ErrorPos -> (Pos,String)
showErrPos' (p,t,msgs) = 
    (p, if p==eotPos
        then unwords ("expected":expected)
	else unwords ("got":string_of_token t: "but expected": expected))
  where
    expected =
      case nub msgs of
	[] -> ["nothing!?"] -- this can't happen
	[e] -> [e]
	es -> "one of":es

updErr :: ErrorPos -> (PosTOKEN,String) -> ErrorPos
updErr e1@(p1,t1,msgs) ((p2,t2),msg) =
    if p1==p2
    then (p1,t1,msg:msgs)
    else if p2<p1
	 then e1
	 else (p2,t2,[msg])


-- general parser

parse :: Parse a -> [PosTOKEN] -> E a
parse p tkl = case parse' p tkl of
	        Left err -> errorE (showErrPos err)
		Right a -> return a

parse' :: Parse a -> [PosTOKEN] -> Either (Pos,String) a
parse' p tkl =
  let (ep,e) = funOfParse (p `chkP` eofP) ((0,0),Symbol "BOT",[]) tkl 
  in elimE e (\(a,tkl) ->
               case tkl of
	         [] -> Right a
	         _  -> Left (showErrPos' ep))
	     (\_ -> Left (showErrPos' ep))

-- Auxiliar

nub :: (Eq a) => [a] -> [a]
nub [] = []
nub (x:xs) = x : nub (filter (/= x) xs)

