module Pretty (Doc, Mode(..), Style(..),
	       empty, text,
	       (<>), hcat, (<+>), (<#>), hsep, ($$), ($$$), vcat,
	       sep, cat, fsep, fcat, nest, alt,
	       renderStyle, fullRender,
	       pretty,
               parens,char,braces,colon,comma)

where

infixl 6 <> 
infixl 6 <+>, <#> -- , <$>
infixl 5 $$, $$$

empty  			  :: Doc
text			  :: String -> Doc 

char  			  :: Char -> Doc
semi, comma, colon, space :: Doc
parens, brackets, braces  :: Doc -> Doc 
quotes, doubleQuotes	  :: Doc -> Doc

(<>)   :: Doc -> Doc -> Doc	-- Beside
hcat   :: [Doc] -> Doc		-- List version of <>
(<+>)  :: Doc -> Doc -> Doc	-- Beside, separated by space
hsep   :: [Doc] -> Doc		-- List version of <+>

($$)   :: Doc -> Doc -> Doc 	-- Above; if there is no
				-- overlap it "dovetails" the two
($$$)  :: Doc -> Doc -> Doc	-- Like $$ but removes superfluous newlines
vcat   :: [Doc] -> Doc		-- List version of $$

cat    :: [Doc] -> Doc		-- Either hcat or vcat
sep    :: [Doc] -> Doc		-- Either hsep or vcat
fcat   :: [Doc] -> Doc		-- ``Paragraph fill'' version of cat
fsep   :: [Doc] -> Doc		-- ``Paragraph fill'' version of sep

nest   :: Int -> Doc -> Doc	-- Nested
						
pretty     :: Doc -> String		-- Uses default style
fullRender :: Mode
	   -> Int			-- Line length
	   -> Float			-- Ribbons per line
	   -> Doc
	   -> String			-- Append this to the end
	   -> String


renderStyle  :: Style -> Doc -> String
data Style = Style { lineLength     :: Int,	-- In chars
		     ribbonsPerLine :: Float,	-- ribbon length / line length
		     mode :: Mode
	     }

style :: Style		-- The default style
style = Style { lineLength = 80, ribbonsPerLine = 2.5, mode = PageMode }

data Mode = PageMode 		-- Normal 
	  | ZigZagMode		-- With zig-zag cuts
	  | LeftMode		-- No indentation
	  | OneLineMode		-- All on one line

char c = text [c]
semi  = char ';'
colon = char ':'
comma = char ','
space = char ' '

quotes p	= char '`' <> p <> char '\''
doubleQuotes p	= char '"' <> p <> char '"'
parens p	= char '(' <> p <> char ')'
brackets p	= char '[' <> p <> char ']'
braces p	= char '{' <> p <> char '}'

hcat = foldr (<>)  empty
hsep = foldr (<#>) empty
vcat = foldr ($$)  empty

p1 <+> p2 = p1 <> space <> p2

p1 `alt` p2 = p1 `Union` p2

data Doc
 = Empty			-- empty
 | NilAbove Doc			-- text "" $$ x
 | TextBeside [Char] Int Doc	-- text s <> x	
 | Nest Int Doc			-- nest k x
 | Union Doc Doc		-- ul U ur
 | NoDoc			-- The empty set of pretties

empty = Empty

text s = case length s of {sl -> TextBeside s sl Empty}

nest 0 p = p			-- Worth a try!
nest k p = Nest k p

p1 $$ p2 = aboveNest p1 0 p2

Empty $$$ y          = y
x     $$$ Empty	     = x
x     $$$ NilAbove y = x $$$ y
x     $$$ y	     = x $$ y

aboveNest :: Doc -> Int -> Doc -> Doc
-- Specfication: aboveNest p k q = p $$ (nest k q)
aboveNest NoDoc            k q = NoDoc
aboveNest Empty               k q = nest k q
aboveNest (NilAbove p)        k q = NilAbove (aboveNest p k q)
aboveNest (TextBeside s sl p) k q = TextBeside s sl (aboveNest1 p (k - sl) q)
aboveNest (Nest k1 p)         k q = Nest k1 (aboveNest p (k - k1) q)
aboveNest (p1 `Union` p2)     k q = aboveNest p1 k q `Union` 
				    aboveNest p2 k q

-- Specification: text s <> aboveNest1 p k q 
--		= text s <> ((text "" <> p) $$ nest k q)
aboveNest1 (Nest _ p) k q = aboveNest1 p k q
aboveNest1 Empty      k q = nilAboveNest k q
aboveNest1 p          k q = aboveNest p k q
	-- All other p's guaranteed non-empty

-- Specification: text s <> nilAboveNest k q 
--		= text s <> (text "" $$ nest k q)
nilAboveNest k Empty = Empty	-- Here's why the "text s <>" is in the spec!
nilAboveNest k (Nest k1 q) = nilAboveNest  (k + k1) q
nilAboveNest k NoDoc = NoDoc
nilAboveNest k q | k > 0		-- No newline if no overlap
		 = TextBeside (spaces k) k q
		 | otherwise		-- Put them really above
		 = NilAbove (nest k q)

NoDoc               <> q = NoDoc
Empty               <> p = deNest p
(NilAbove p)        <> q = NilAbove (p <> q)
(TextBeside s sl p) <> q = TextBeside s sl (p <> q)
(Nest k p)          <> q = Nest k (p <> q)
(p1 `Union` p2)     <> q = (p1 <> q) `Union` (p2 <> q)

deNest (Nest _ p) = p
deNest p          = p

NoDoc      <#> q = NoDoc
Empty	   <#> q = q
(Nest k p) <#> q = Nest k (p <#> q)
p 	   <#> q = p <> prefixSP q	-- p non-empty

prefixSP :: Doc -> Doc
prefixSP NoDoc = NoDoc
prefixSP (Nest k p) = prefixSP p
prefixSP Empty = Empty
prefixSP p = TextBeside " " 1 p

-- Specification: sep ps  = oneLiner (hsep ps)
--			   `Union`
--			    vcat ps

sep = sepX True		-- Separate with spaces
cat = sepX False		-- Don't

sepX x []     = empty
sepX x (p:ps) = sep1 x p 0 ps

-- Specification: sep1 x k ys = sep (x : map (nest k) ys)
--			      = oneLiner (x <#> nest k (hsep ys))
--				`Union` x $$ nest k (vcat ys)

sep1 x NoDoc 		   k ys = NoDoc
sep1 x Empty               k ys = nest k (sepX x ys)
sep1 x (NilAbove p)        k ys = NilAbove (aboveNest p k (vcat ys))
sep1 x (TextBeside s sl p) k ys = TextBeside s sl (sepNB x p (k - sl) ys)
sep1 x (Nest n p)          k ys = Nest n (sep1 x p (k - n) ys)
sep1 x (p `Union` q)       k ys = sep1 x p k ys
				`Union`
				 (q $$ nest k (vcat ys))

-- Specification: sepNB p k ys = sep1 (text "" <> p) k ys
-- Called when we have already found some text in the first item

sepNB x NoDoc k ys	  = NoDoc
sepNB x Empty k ys        = oneLiner (if x then prefixSP (hsep ys)
					   else hcat ys)
			    `Union` 
			    nilAboveNest k (vcat ys)
sepNB x (Nest _ p)  k ys  = sepNB x p k ys
sepNB x p k ys		  = sep1 x p k ys

fsep = fill True
fcat = fill False

-- Specification: fill []  = empty
--		   fill (p:ps) = oneLiner p <#> oneLiner (hsep ps)
--			   `Union`
--			    p $$ fill ps

fill x []     = empty
fill x (p:ps) = fill1 x p 0 ps


-- Specification: sep1 x k ys = sep (x : map (nest k) ys)
--			      = oneLiner (x <#> nest k (hsep ys))
--				`Union` x $$ nest k (vcat ys)
--
fill1 x NoDoc	 	    k ys = NoDoc
fill1 x Empty               k ys = nest k (fill x ys)
fill1 x (NilAbove p)        k ys = NilAbove (aboveNest p k (fill x ys))
fill1 x (TextBeside s sl p) k ys = TextBeside s sl (fillNB x p (k - sl) ys)
fill1 x (Nest n p)          k ys = Nest n (fill1 x p (k - n) ys)
fill1 x (p `Union` q)       k ys = fill1 x p k ys
				   `Union`
				   (q $$ nest k (fill x ys))

fillNB x NoDoc k ys	   = NoDoc
fillNB x Empty k []        = Empty
fillNB x Empty k (y:ys)    = (if x then prefixSP (fill1 x (oneLiner y) (k - 1) ys)
			           else           fill1 x (oneLiner y) k         ys)
			     `Union` 
			     nilAboveNest k (fill x (y:ys))
fillNB x (Nest _ p)  k ys  = fillNB x p k ys
fillNB x p k ys		   = fill1 x p k ys

best :: Mode
     -> Int		-- Line length
     -> Int		-- Ribbon length
     -> Doc
     -> Doc		-- No Unions in here!

best OneLineMode w r p
  = get p
  where
    get Empty               = Empty
    get NoDoc               = NoDoc
    get (NilAbove p)        = NilAbove (get p)
    get (TextBeside s sl p) = TextBeside s sl (get p)
    get (Nest k p)          = get p		-- Elide nest
    get (p `Union` q)       = first (get p) (get q)

best mode w r p
  = get w p
  where
    get :: Int		-- (Remaining) width of line
        -> Doc -> Doc
    get w Empty               = Empty
    get w NoDoc               = NoDoc
    get w (NilAbove p)        = NilAbove (get w p)
    get w (TextBeside s sl p) = TextBeside s sl (get1 w sl p)
    get w (Nest k p)          = Nest k (get (w - k) p)
    get w (p `Union` q)       = nicest w r (get w p) (get w q)

    get1 :: Int		-- (Remaining) width of line
         -> Int		-- Amount of first line already eaten up
         -> Doc		-- This is an argument to TextBeside => eat Nests
         -> Doc		-- No Unions in here!

    get1 w sl Empty               = Empty
    get1 w sl NoDoc               = NoDoc
    get1 w sl (NilAbove p)        = NilAbove (get (w - sl) p)
    get1 w sl (TextBeside t tl p) = TextBeside t tl (get1 w (sl + tl) p)
    get1 w sl (Nest k p)          = get1 w sl p
    get1 w sl (p `Union` q)       = nicest1 w r sl (get1 w sl p) 
				                   (get1 w sl q)

nicest w r p q = nicest1 w r 0 p q
nicest1 w r sl p q | fits ((w `minn` r) - sl) p = p
		   | otherwise 	                 = q

fits :: Int	-- Space available
     -> Doc
     -> Bool	-- True if *first line* of Doc fits in space available
 
fits n p    | n < 0 = False
fits n NoDoc               = False
fits n Empty               = True
fits n (NilAbove _)        = True
fits n (TextBeside _ sl p) = fits (n - sl) p

minn x y | x < y    = x
	 | otherwise = y

first p q | nonEmptySet p = p 
	  | otherwise     = q

nonEmptySet NoDoc           = False
nonEmptySet (p `Union` q)      = True
nonEmptySet Empty	       = True
nonEmptySet (NilAbove p)       = True		-- NoDoc always in first line
nonEmptySet (TextBeside _ _ p) = nonEmptySet p
nonEmptySet (Nest _ p)         = nonEmptySet p

oneLiner :: Doc -> Doc
oneLiner NoDoc               = NoDoc
oneLiner Empty               = Empty
oneLiner (NilAbove p)        = NoDoc
oneLiner (TextBeside s sl p) = TextBeside s sl (oneLiner p)
oneLiner (Nest k p)          = Nest k (oneLiner p)
oneLiner (p `Union` pq)      = oneLiner p

renderStyle Style{mode, lineLength, ribbonsPerLine} doc 
  = fullRender mode lineLength ribbonsPerLine doc ""

pretty doc = renderStyle style doc

fullRender mode line_length ribbons_per_line doc cont
  = display mode line_length ribbon_length best_doc cont
  where 
    best_doc = best mode hacked_line_length ribbon_length doc

    hacked_line_length, ribbon_length :: Int
    ribbon_length = round (fromInt line_length / ribbons_per_line)
    hacked_line_length =
        case mode of { ZigZagMode -> maxBound; other -> line_length }

display OneLineMode _ _ doc cont = easy_display ' '  doc cont
display LeftMode    _ _ doc cont = easy_display '\n' doc cont

display mode page_width ribbon_width doc cont
  = case page_width - ribbon_width of { gap_width ->
    case gap_width `quot` 2 of { shift ->
    let
    	lay k (Nest k1 p)  = lay (k + k1) p
    	lay k Empty        = cont
    
    	lay k (NilAbove p) = '\n' : lay k p
    
    	lay k (TextBeside s sl p)
	    = case mode of
		    ZigZagMode |  k >= gap_width
			       -> '\n' : multi_ch shift '/' (
				  '\n' : lay1 (k - shift) s sl p)

			       |  k < 0
			       -> '\n' : multi_ch shift '\\' (
				  '\n' : lay1 (k + shift) s sl p )

		    other -> lay1 k s sl p
    
    	lay1 k s sl p = indent k (s ++ (lay2 (k + sl) p))
    
    	lay2 k (NilAbove p)        = '\n' : lay k p
    	lay2 k (TextBeside s sl p) = s ++ (lay2 (k + sl) p)
    	lay2 k (Nest _ p)          = lay2 k p
    	lay2 k Empty               = cont
    in
    lay 0 doc
    }}

easy_display nl_char doc cont 
  = lay doc
  where
    lay (Nest k p)   	    = lay p
    lay Empty        	    = cont
    lay (NilAbove p) 	    = nl_char : lay p
    lay (TextBeside s sl p) = s ++ lay p

indent n rest | n >= 8 = '\t' : indent (n - 8) rest
	      | otherwise = multi_ch n ' ' rest

multi_ch 0 ch rest = rest
multi_ch n       ch rest = ch : multi_ch (n - 1) ch rest

spaces 0 = ""
spaces n       = ' ' : spaces (n - 1)


